# Copyright (C) 2023-2024 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from __future__ import annotations
from typing import cast
from typing import Any
from typing import ClassVar
from typing import Generic
from typing import Self

import pydantic
import pydantic.fields
from pydantic.fields import FieldInfo

from canonical.ext.crypto import EncryptionResult
from canonical.ext.crypto import IDataEncryption
from canonical.lib.types import ResourceName
from .apimodelfieldinfo import APIModelFieldInfo
from .apimodelinspector import APIModelInspector
from .builders import APIModelClassBuilder


__all__: list[str] = [
    'APIModel'
]


class APIModel(pydantic.BaseModel):
    inspector: ClassVar[APIModelInspector] = APIModelInspector()
    model_config = {'populate_by_name': True}
    __builder__: ClassVar[APIModelClassBuilder[Any] | None] = None
    __builder_class__: ClassVar[type[APIModelClassBuilder[Any]]] = APIModelClassBuilder
    __create_model__: ClassVar[type[pydantic.BaseModel]]
    __update_model__: ClassVar[type[pydantic.BaseModel]]
    __stored_model__: ClassVar[type[Self]]
    __output_model__: ClassVar[type[pydantic.BaseModel]]
    __origin_fields__: ClassVar[dict[str, APIModelFieldInfo]] = {}
    __metadata_class__: ClassVar[type[pydantic.BaseModel] | None] = None
    __mode__: ClassVar[str] = 'domain'
    __namespaced__: ClassVar[bool] = False
    __params__: ClassVar[dict[str, Any]]
    _dirty: bool = pydantic.PrivateAttr(default=False)
    _parent: pydantic.BaseModel | None = pydantic.PrivateAttr(default=None)

    @classmethod
    def build_class(cls, builder: Any, **kwargs: Any):
        builder = cast(APIModelClassBuilder['APIModel'], builder)
        builder.contribute_annotations()

    @classmethod
    def contribute_to_class(
        cls,
        parent: type[APIModel],
        attname: str,
        field: pydantic.fields.FieldInfo
    ) -> None:
        """Hook to execute when this model is an attribute on
        a new class.
        """
        pass

    @classmethod
    def model_input(cls, data: dict[str, Any]):
        return cls.__create_model__.model_validate(data)

    @classmethod
    def model_input_json(cls, buf: str | bytes):
        return cls.__create_model__.model_validate_json(buf)

    @classmethod
    def __pydantic_init_subclass__(cls, **kwargs: Any) -> None:
        if not cls.inspector.is_concrete(cls):
            return
        origin = cls.__pydantic_generic_metadata__.get('origin')
        if kwargs.get('abstract') or origin is not None:
            return
        if kwargs:
            cls.__namespaced__ = kwargs.pop('namespaced', cls.__namespaced__)
        cls.__builder__ = cls.__builder_class__(
            model=cls,
            metadata_class=cls.__metadata_class__,
            namespaced=cls.__namespaced__,
            **kwargs
        )
        cls.build_class(cls.__builder__, **kwargs)
        cls.__builder__.construct()

    def __init_subclass__(cls, **kwargs: Any):
        for Base in cls.__bases__:
            if Base in (pydantic.BaseModel, Generic):
                super().__init_subclass__()
                continue
            #super().__init_subclass__(**kwargs)

    @property
    def parent(self) -> APIModel | None:
        return cast(APIModel, self._parent)

    def add_to_parent(self, parent: Any):
        self._parent = parent

    def get_resource_name(self) -> ResourceName:
        """Return the :class:`ResourceName` where the
        resource may be found."""
        raise NotImplementedError

    def get_url(self) -> str:
        """Return the URL where the resource may be found."""
        raise NotImplementedError

    def is_attached(self) -> bool:
        return self._parent is not None

    def model_post_init(self, _: Any) -> None:
        for attname in self.model_fields:
            v = getattr(self, attname)
            if not isinstance(v, APIModel):
                continue
            v.add_to_parent(self)

    def must_encrypt(self, attname: str) -> bool:
        return False

    def replace(self, instance: pydantic.BaseModel):
        """Replaces all fields of `self` with corresponding fields of
        `instance.
        """
        for attname, field in self.model_fields.items():
            if attname not in instance.model_fields:
                continue
            adapter = pydantic.TypeAdapter(field.annotation)
            ours = getattr(self, attname)
            theirs = cleaned = getattr(instance, attname, None)
            if not isinstance(theirs, pydantic.BaseModel):
                cleaned = adapter.validate_python(theirs)
            if not field.frozen:
                # TODO: must actually raise an exception if the value is
                # different, but it needs to integrate with FastAPIs
                # request validation.
                if isinstance(cleaned, pydantic.BaseModel) and isinstance(ours, APIModel):
                    ours.replace(cleaned)
                    continue
                self.replace_field(attname, ours, theirs)
        return self

    def replace_field(self, attname: str, ours: Any, theirs: Any):
        if ours != theirs:
            setattr(self, attname, theirs)

    async def encrypt(self, dek: IDataEncryption[Any]) -> None:
        dek = await self.get_data_encryption(dek)
        for attname, field in self.model_fields.items():
            await self.encrypt_field(dek, field, attname, getattr(self, attname))

    async def encrypt_field(
        self,
        dek: IDataEncryption[Any],
        field: APIModelFieldInfo | FieldInfo,
        attname: str,
        value: Any
    ) -> None:
        assert field.annotation is not None
        if value is None:
            return

        if isinstance(value, EncryptionResult):
            return

        if isinstance(value, APIModel):
            await value.encrypt(dek)
            return

        if (isinstance(field, APIModelFieldInfo) and field.encrypt)\
        or self.must_encrypt(attname):
            setattr(self, attname, await dek.encrypt(value, field.annotation))
            return

    async def decrypt(self, dek: IDataEncryption[Any]):
        dek = await self.get_data_encryption(dek)
        for attname in self.model_fields.keys():
            value = getattr(self, attname)
            if isinstance(value, EncryptionResult):
                setattr(self, attname, await dek.decrypt(value))
            if isinstance(value, APIModel):
                await value.decrypt(dek)

    async def get_data_encryption(
        self,
        dek: IDataEncryption[Any]
    ) -> IDataEncryption[Any]:
        return dek


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any


class BaseMetaclass(type):

    def __new__(
        cls,
        name: str,
        bases: tuple[type, ...],
        attrs: dict[str, Any],
        **kwargs: Any
    ):
        attrs.setdefault('handle', cls.handle)
        return super().__new__(cls, name, bases, attrs, **kwargs)

    def handle(cls) -> None:
        raise NotImplementedError


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Generic
from typing import Literal
from typing import TypeVar

import pydantic

from .listmeta import ListMeta


K = TypeVar('K')
R = TypeVar('R')


class ListBase(pydantic.BaseModel, Generic[K, R]):
    """List base class."""
    model_config = {'populate_by_name': True}

    api_version: Literal['v1'] = pydantic.Field(
        default=...,
        alias='apiVersion',
        title="API Version",
        description=(
            "The `apiVersion` field defines the versioned schema of this "
            "representation of an object. Servers should convert recognized "
            "schemas to the latest internal value, and may reject "
            "unrecognized values."
        )
    )

    kind: K = pydantic.Field(
        default=...,
        description=(
            "Kind is a string value representing the REST resource this "
            "object represents. Servers may infer this from the endpoint "
            "the client submits requests to. Cannot be updated. In `CamelCase`."
        )
    )

    metadata: ListMeta = pydantic.Field(
        default=...,
        description="Standard list metadata."
    )

    items: list[R] = []


# Copyright (C) 2023-2024 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from __future__ import annotations
from typing import Any
from typing import TypeVar
from typing import TYPE_CHECKING

import pydantic

from .apimodel import APIModel
from .bases import BaseReference
from .fields import APIVersion
from .fields import Kind
from .fields import Name
from .fields import Namespace
if TYPE_CHECKING:
    from .objectmeta import ObjectMeta

__all__: list[str] = [
    'ObjectReference'
]

T = TypeVar('T')


class ObjectReference(APIModel, BaseReference):
    """An `ObjectReference` instance contains enough information
    to let you inspect or modify the referred object.
    """
    model_config = {'populate_by_name': True}

    api_version: APIVersion = pydantic.Field(
        default=...,
        description=(
            "Specifies the API version of the referent. Cannot be updated."
        ),
        frozen=True
    )

    kind: Kind = pydantic.Field(
        default=...,
        description=(
            "Kind of the referent. Cannot be updated. In `CamelCase`."
        ),
        frozen=True
    )

    name: Name = pydantic.Field(
        default=...,
        description=(
            "The `.metadata.name` of of the referent. Cannot be "
            "updated."
        ),
        frozen=True
    )

    namespace: Namespace = pydantic.Field(
        default='',
        description="The namespace of the referent.",
        frozen=False
    )

    resource_version: str = pydantic.Field(
        default='',
        description="Specific resourceVersion to which this reference is made, if any."
    )

    uid: int | None = pydantic.Field(
        default=None,
        description="UID of the referent."
    )

    @property
    def api_group(self):
        return self.api_version.group

    def as_name(self) -> str:
        return self.name

    def attach(self, metadata: ObjectMeta[Any]):
        self.resource_version = getattr(metadata, 'resource_version', '')
        self.uid = getattr(metadata, 'uid', None)
        self.namespace = getattr(metadata, 'namespace', '')
        return self

    def cache_key(self, prefix: str):
        if self.namespace:
            prefix = f'{prefix}:{self.namespace}'
        return f'{prefix}:{self.name}'

    def get_namespace(self) -> str | None:
        return self.namespace or None

    def is_cluster(self) -> bool:
        return not bool(self.namespace)

    def in_namespace(self, namespace: str) -> bool:
        return self.namespace == namespace

    def is_local(self) -> bool:
        return not self.is_cluster()

    def is_namespaced(self) -> bool:
        return bool(self.namespace)

    def with_namespace(self, namespace: str):
        if namespace != self.namespace:
            raise ValueError(
                f"{type(self).__name__}.namespace '{self.namespace}' is not "
                f"equal to '{namespace}'."
            )
        return self


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic
import pytest

from canonical.ext.api import APIResourceModel
from canonical.ext.api import ObjectMeta


class Order(APIResourceModel[int], group='', version='v2', plural='rs', namespaced=True):
    model_config = {'extra': 'forbid'}


def test_default_implementation_is_not_namespaced():
    assert 'namespace' not in ObjectMeta.model_fields


def test_name_is_validated():
    with pytest.raises(pydantic.ValidationError):
        Order.model_input({
            'metadata': {'namespace': 'foo', 'name': 'bar'}
        })

    Order.model_input({
        'metadata': {'namespace': 'foo', 'name': 1}
    })


# Copyright (C) 2023-2024 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic

from .apimodel import APIModel
from .bases import BaseReference
from .types import APIVersion


__all__: list[str] = [
    'UIDReference'
]


class UIDReference(APIModel, BaseReference):
    _namespace: str | None = pydantic.PrivateAttr(default=None)
    api_version: APIVersion
    kind: str
    uid: int

    @property
    def api_group(self) -> str:
        return self.api_version.group

    def as_name(self) -> int | str:
        return self.uid

    def get_namespace(self) -> str | None:
        return None


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import TypeVar

from canonical.ext.api import APIResourceType
from ..params import RequestResource
from .default import Default


T = TypeVar('T', bound=APIResourceType)


class Retrieve(Default[T]):
    detail = True
    existing = True
    method = 'GET'
    status_code = 200
    verb = 'get'


    async def handle(self, obj: RequestResource) -> Any:
        return obj

    def get_endpoint_summary(self) -> str:
        return f'Retrieve a specific {self.model.__name__}'

    def get_openapi_responses(self) -> dict[int, dict[str, Any]]:
        return {
            **super().get_openapi_responses(),
            200: {
                'model': self.model,
                'description': self.get_response_description()
            }
        }

    def get_response_description(self) -> str:
        return f"{self.model.__name__} object."


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import logging
from typing import Any
from typing import TypeVar

import pydantic

from ..models import EncryptionResult
from ..protocols import IDataEncryption


T = TypeVar('T')


class BaseDEK(IDataEncryption[EncryptionResult]):
    __module__: str = 'canonical.ext.crypto.bases'
    logger: logging.Logger = logging.getLogger('uvicorn')

    async def decrypt(
        self,
        ct: EncryptionResult,
        decoder: type[T] = bytes
    ) -> T:
        self.logger.debug("Decrypting (encoder: %s)", repr(decoder))
        adapter: pydantic.TypeAdapter[T] = pydantic.TypeAdapter(decoder)
        pt = await self.perform_decrypt(ct)
        return adapter.validate_json(pt)

    async def encrypt(
        self,
        value: Any,
        encoder: type = bytes
    ) -> EncryptionResult:
        if isinstance(value, EncryptionResult):
            # Only here to suppress warnings (TODO).
            return value
        self.logger.debug("Encrypting (encoder: %s)", repr(encoder))
        adapter: pydantic.TypeAdapter[bytes] = pydantic.TypeAdapter(encoder)
        buf = value
        if not isinstance(value, bytes):
            buf = adapter.dump_json(value, warnings=False)
        assert isinstance(buf, bytes)
        return await self.perform_encrypt(buf)

    async def perform_decrypt(self, ct: EncryptionResult) -> bytes:
        raise NotImplementedError

    async def perform_encrypt(self, value: bytes) -> EncryptionResult:
        raise NotImplementedError


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .authorizationserverendpoint import AuthorizationServerEndpoint
from .baseendpoint import BaseEndpoint
from .client import BaseClient
from .provider import BaseProvider
from .providerspec import BaseProviderSpec


__all__: list[str] = [
    'AuthorizationServerEndpoint',
    'BaseClient',
    'BaseEndpoint',
    'Endpoint',
    'BaseProvider',
    'BaseProviderSpec',
]

Endpoint = BaseEndpoint


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Union

import httpx
import pydantic

from canonical.ext.api import APIResourceModel
from canonical.ext.oauth.models.responses import AuthorizationResponse
from canonical.ext.oauth.models.responses import TokenResponse
from canonical.ext.oauth.models.requests import AuthorizationCodeRequest
from canonical.ext.oauth.protocols import IObtainingClient
from .authorizationstatestatus import AuthorizationRequestStateStatus
from .forwardingstatespec import ForwardingStateSpec
from .issuingstatespec import IssuingStateSpec
from .obtainingstatespec import ObtainingStateSpec
from .proxyingstatespec import ProxyingStateSpec


AuthorizationStateSpecType = Union[
    ForwardingStateSpec,
    IssuingStateSpec,
    ObtainingStateSpec,
    ProxyingStateSpec,
]


class AuthorizationState(
    APIResourceModel[str],
    group='oauth',
    version='v2',
    plural='authorizationstates'
):
    """Describes the state of an OAuth 2.x/OpenID Connect
    authorization request.
    """
    spec: AuthorizationStateSpecType = pydantic.Field(
        default=...,
        title="Specification",
        description=(
            "Specifies the parameters of the authorization request."
        )
    )

    status: AuthorizationRequestStateStatus = pydantic.Field(
        default=...,
        description=(
            "Describes the current state of the AuthorizationRequest."
        )
    )

    def get_user_agent_redirect_uri(self) -> str:
        return self.spec.get_user_agent_redirect_uri(self.metadata)

    def is_failed(self) -> bool:
        return self.status.current == 'Failed'

    def is_forwarding(self) -> bool:
        return self.spec.redirect_mode == 'forwarding'

    def is_obtained(self) -> bool:
        return self.status.current == 'Obtained'

    def is_obtainable(self) -> bool:
        return self.spec.redirect_mode in {'issuing', 'proxy'}

    def is_proxy(self) -> bool:
        return self.spec.redirect_mode == 'proxy'

    async def obtain(
        self,
        http: httpx.AsyncClient,
        client: IObtainingClient[TokenResponse],
        params: AuthorizationResponse
    ):
        assert params.code is not None
        client = client.configure(
            server=self.spec.issuer,
            token_endpoint=self.spec.token_endpoint
        )
        response = await client.obtain(
            grant=AuthorizationCodeRequest(
                grant_type='authorization_code',
                code=params.code,
                redirect_uri=self.spec.get_redirect_uri()
            ),
            http=http
        )
        if not response.is_error():
            assert isinstance(response, TokenResponse)
            self.status.update(self.metadata, **self.spec.on_obtained(response))
        else:
            self.status.update(
                metadata=self.metadata,
                status='Failed'
            )
        return response

    def get_return_url(self, iss: str) -> str:
        """Get the URL to redirect the user agent after succesfully
        obtaining an access code.
        """
        return self.spec.get_return_url(
            code=self.metadata.name,
            iss=iss
        )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from __future__ import annotations
import datetime
from typing import Any
from typing import TYPE_CHECKING

import pydantic

if TYPE_CHECKING:
    from .resource import Resource
from .resourcekey import ResourceKey
from .resourcemetadata import ResourceMetadata


class TransientMeta(ResourceMetadata):
    model_config = {'populate_by_name': True}
    
    created: datetime.datetime | None = pydantic.Field(
        default_factory=lambda: datetime.datetime.now(datetime.timezone.utc),
        title="Created",
        description=(
            "The `created` field represents the date and time at which this "
            "object was created."
        ),
        frozen=True
    )

    @classmethod
    def is_namespaced(cls) -> bool:
        return False

    @classmethod
    def default(cls):
        return cls()

    @property
    def name(self) -> str:
        raise NotImplementedError

    @property
    def key(self) -> ResourceKey[Any]:
        raise NotImplementedError

    @property
    def resource_version(self) -> None:
        return None

    @classmethod
    def add_to_model(cls, model: type[Resource[Any]]) -> None:
        model.model_fields['metadata'].default = None
        model.model_fields['metadata'].default_factory = cls.default

    @classmethod
    def build_input_model(cls):
        return cls

    def attach(self, resource: Resource[Any]):
        self._api_version = resource.api_version
        self._kind = resource.kind

    def get_namespace(self) -> str | None:
        return None

    def update_resource_version(self, obj: Resource[Any]):
        pass

    @staticmethod
    def get_metadata_field():
        return pydantic.Field(
            default=None,
            title='Metadata',
            description=(
                "`ObjectMeta` is metadata that all persisted resources "
                "must have, which includes all objects users must create."
            )
        )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import TypeVar

from .bases import BaseDEK


T = TypeVar('T')


class NullDataEncryption(BaseDEK):
    __module__: str = 'canonical.ext.crypto'

    async def decrypt(self, ct: T, decoder: type[T] = bytes) -> T:
        self.logger.critical(
            "Decryption is not implemented, returning %s.",
            type(ct).__name__
        )
        return ct

    async def encrypt(self, value: Any, encoder: type = bytes) -> Any:
        self.logger.critical(
            "Encryption is not implemented, returning %s.",
            type(value).__name__
        )
        return value


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Annotated
from typing import TypeAlias

import fastapi

from canonical.ext.resource import ObjectReference
from canonical.ext.resource import ResourceInspector
from .resourcemodel import ResourceModel


inspector = ResourceInspector()


def get(
    request: fastapi.Request,
    model: ResourceModel
):
    state = request.state
    if not hasattr(state, 'name'):
        return None
    meta = inspector.inspect(model)
    ref = ObjectReference(
        api_version=meta.api_version,
        kind=meta.kind,
        name=getattr(state, 'name'),
        namespace=getattr(state, 'namespace', '')
    )
    return ref.with_model(model)


RequestObjectReference: TypeAlias = Annotated[
    ObjectReference | None,
    fastapi.Depends(get)
]


# Copyright (C) 2023-2024 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Literal
from typing import TypeVar

from canonical.ext.cache import NullCache
from canonical.lib.protocols import ICache
from .baseresourcerepository import BaseResourceRepository
from .resourceinspector import ResourceInspector
from .resourcekey import ResourceKey


T = TypeVar('T')


class ResourceResolver:
    """Resolve a :class:`~canonical.ext.api.APIResource` instance by a
    supported key type.
    """
    inspector: ResourceInspector = ResourceInspector()

    def __init__(
        self,
        repo: BaseResourceRepository,
        cache: ICache = NullCache()
    ):
        self.cache = cache
        self.repo = repo

    def generate_cache_key(self, model: type[Any], key: ResourceKey) -> str:
        """Return a string that may be used as a cache key based on the
        key of a resource.
        """
        return self.inspector.cache_key(model, key)

    async def from_cache(self, key: str, model: type[T]) -> T | None:
        """Lookup a resource from the cache using the given cache
        key `key`. If there is no cached version of the resource,
        return ``None``.
        """
        return await self.cache.get(key, model)

    async def lookup(
        self,
        key: ResourceKey,
        model: type[T]
    ):
        """Lookup a resource by its key from its master data store."""
        return await self.repo.get(key, model)

    async def resolve(
        self,
        key: ResourceKey,
        model: type[T],
        *,
        ttl: int | None = None,
        mode: Literal['fresh', 'cache-first'] = 'cache-first'
    ) -> T:
        """Resolve a resource by its key using the given `mode`.

        If mode is ``fresh``, force a lookup from the master data
        store of the resource, otherwise attempt to retrieve it
        from the cache first.

        The `ttl` is an integer that specifies the number seconds
        that a freshly retrieved resource should be maintained in
        the cached. If `ttl` is ``None``, then it is up to the
        cache backend to decide on if and how long to cache the
        resource.
        """
        cache_key = self.generate_cache_key(model, key)
        instance = None
        if mode == 'cache-first':
            instance = await self.from_cache(cache_key, model)
        return await self.on_resolved(
            model=model,
            instance=instance or await self.lookup(key, model),
            cache_key=cache_key,
            fresh=instance is None,
            ttl=ttl
        )

    async def on_resolved(
        self,
        model: type[T],
        instance: T,
        *,
        cache_key: str,
        fresh: bool,
        ttl: int | None
    ) -> T:
        if fresh:
            await self.cache.set(cache_key, instance, model, ttl=ttl)
        return instance


# Copyright (C) 2023-2024 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Annotated

import pydantic

from .types import APIVersion as APIVersionType


__all__: list[str] = [
    'APIVersion',
    'Kind'
]

APIVersion = Annotated[
    APIVersionType,
    pydantic.Field(
        alias='apiVersion',
        title="API Version",
        description=(
            "The `apiVersion` field defines the versioned schema of this "
            "representation of an object. Servers should convert recognized "
            "schemas to the latest internal value, and may reject "
            "unrecognized values."
        ),
        frozen=True
    )
]

Kind = Annotated[
    str,
    pydantic.Field(
        description=(
            "Kind is a string value representing the REST resource this "
            "object represents. Servers may infer this from the endpoint "
            "the client submits requests to. Cannot be updated. In `CamelCase`."
        ),
        frozen=True
    )
]

Name = Annotated[
    str,
    pydantic.Field(
        description=(
            "Name must be unique within a namespace. Is required when creating "
            "resources, although some resources may allow a client to request "
            "the generation of an appropriate name automatically. Name is "
            "primarily intended for creation idempotence and configuration"
            "definition."
        )
    )
]

Namespace = Annotated[
    str,
    pydantic.Field(
        frozen=True
    )
]

ReferencedName = Annotated[
    str,
    pydantic.Field(
        description=(
            "The `name` field is the name of resource being "
            "referenced"
        )
    )
]

UID = Annotated[
    int,
    pydantic.Field(
        title="UID",
        description=(
            "UID is the unique in time and space value for this object. "
            "It is typically generated by the server on successful creation "
            "of a resource and is not allowed to change on PUT operations. "
            "Populated by the system. Read-only."
        )
    ),
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

from .dependencyconfig import DependencyConfig
from .requirement import Requirement
from .provider import inject
from .provider import provide


__all__: list[str] = [
    'configure',
    'inject',
    'Require'
]


def Require(name: str) -> Any:
    return Requirement(name)


def configure(fn: str):
    config = DependencyConfig.open(fn)
    for dep in config.dependencies:
        provide(dep)


# Copyright (C) 2023-2024 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Generic
from typing import TypeVar

from canonical.lib.protocols import IBuilder


T = TypeVar('T')


class BaseBuilder(IBuilder[T], Generic[T]):

    def __init__(self, cls: type[T], **kwargs: Any):
        self.model = cls
        self.setup(cls, **kwargs)

    def setup(self, model: type[T], **kwargs: Any) -> None:
        raise NotImplementedError

    async def build(self) -> T:
        raise NotImplementedError


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Literal
from inspect import Parameter
from inspect import Signature
from typing import Any
from typing import Callable
from typing import Union

import aorta
import pydantic

from canonical.lib.protocols import ICommand
from .default import Default


class Command(Default[Any]):
    detail = True
    exists = True
    method = 'POST'
    requires_body = True
    verb = 'issueCommand'

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.commands: list[tuple[type[ICommand], str, Callable[..., Any]]] = []

    def add(self, command: type[ICommand], verb: str, handler: Callable[..., Any]) -> None:
        self.commands.append((command, verb, handler))

    def annotate_handler(self, signature: Signature) -> Signature:
        annotation = Union[*[  # type: ignore
            self.build_command_model(c)  # type: ignore
            for c, *_ in self.commands
        ]]
        return signature.replace(
            parameters=[
                Parameter(
                    kind=Parameter.POSITIONAL_ONLY,
                    name='command',
                    annotation=annotation
                ),
                *signature.parameters.values()
            ]
        )

    def build_command_model(self, cls: type[aorta.Command]):
        attrs: dict[str, Any] = {
            'api_version': pydantic.Field(
                default=...,
                alias='apiVersion',
                title='API Version',
                description=(
                    "The `apiVersion` field defines the versioned schema of "
                    "this representation of a command."
                )
            ),
            'kind': pydantic.Field(
                default=...,
                description=(
                    "Kind is a string value representing the kind of command "
                    "that is issued."
                )
            ),
            'spec': pydantic.Field(
                default=...,
                title='Specification',
                description='Specification of the command parameters.'
            ),
            '__annotations__': {
                'api_version': Literal[f'{self.model.__meta__.api_version}'],
                'kind': Literal[f'{cls.__name__}'],
                'spec': cls
            }
        }
        return type(cls.__name__, (pydantic.BaseModel,), attrs)

    def get_endpoint_summary(self) -> str:
        return f'Issue a command'

    def get_response_description(self) -> str:
        return 'IssuedCommand object.'

    def get_path(self) -> str:
        return f'{super().get_path()}/commands'

    async def handle(self) -> None:
        raise NotImplementedError


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Literal

import httpx
import pydantic

from canonical.lib.types import HTTPResourceLocator
from .servermetadata import ServerMetadata


class OIDCProvider(pydantic.BaseModel):
    model_config = {'populate_by_name': True}

    protocol: Literal['oauth2:oidc']

    issuer: HTTPResourceLocator = pydantic.Field(
        default=...,
        description=(
            "The OpenID Connect Issuer Identifier."
        )
    )

    metadata_url: str | None = pydantic.Field(
        default=None,
        alias='metadataUrl',
        description=(
            "The metadata URL used for discovery, if the provider "
            "uses a non-standard path."
        )
    )

    async def discover(self, http: httpx.AsyncClient):
        metadata = None
        urls: list[str] = [
            f'{self.issuer}/.well-known/oauth-authorization-server',
            f'{self.issuer}/.well-known/openid-configuration'
        ]
        if self.metadata_url:
            urls = [self.metadata_url]
        for url in urls:
            response = await http.get(url)
            if response.status_code == 404:
                continue
            response.raise_for_status()
            metadata = ServerMetadata.model_validate(response.json())
            await metadata.discover(http=http)
            break
        if metadata is None:
            raise NotImplementedError
        return metadata


# Copyright (C) 2023-2024 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from canonical.lib.types import StringType


class APIVersion(StringType):
    group: str
    version: str

    def type_post_init(self):
        if self.find('/') != -1:
            self.group, self.version = str.rsplit(self, '/', maxsplit=1)
        else:
            self.group = ''
            self.version = str(self)

    def __repr__(self) -> str:
        return str(self)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import hashlib
from typing import Any
from typing import Generic
from typing import Literal
from typing import TypeVar

from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives.asymmetric.padding import AsymmetricPadding
from cryptography.hazmat.primitives.hashes import HashAlgorithm

from canonical.lib.types import Base64Int
from canonical.lib.utils.encoding import b64encode
from .base import BaseJWK


A = TypeVar('A')
O = TypeVar('O')
U = TypeVar('U')


class RSAKey(
    BaseJWK[
        Literal['RSA'], A, O, U
    ],
    Generic[A, O, U]
):
    e: Base64Int
    n: Base64Int
    _digest: type[HashAlgorithm]
    _public_numbers: rsa.RSAPublicNumbers
    _public_key: rsa.RSAPublicKey

    @property
    def thumbprint(self) -> str:
        message = self.model_dump_json(include={'e', 'kty', 'n'})
        return b64encode(hashlib.sha256(str.encode(message)).digest(), encoder=bytes.decode)

    def model_post_init(self, _: Any) -> None:
        self._public_numbers = rsa.RSAPublicNumbers(
            e=self.e,
            n=self.n
        )
        self._digest = self.get_digest_algorithm()
        self._public_key = self._public_numbers.public_key()

    def get_digest_algorithm(self) -> type[HashAlgorithm]:
        raise NotImplementedError

    def get_padding(self) -> AsymmetricPadding:
        cls = self.get_padding_class()
        return cls(**self.get_padding_params(cls, self._digest))

    def get_padding_class(self) -> type[AsymmetricPadding]:
        raise NotImplementedError

    def get_padding_params(
        self,
        padding_class: type[AsymmetricPadding],
        digest_class: type[HashAlgorithm],
    ) -> dict[str, Any]:
        raise NotImplementedError


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import functools
import json
from typing import Any
from typing import Generic
from typing import TypeVar

import fastapi
import pydantic
import yaml


T = TypeVar('T', bound=pydantic.BaseModel)


class Response(fastapi.Response, Generic[T]):
    content: pydantic.BaseModel
    default_media_type: str | None = None
    model: type[T]

    @classmethod
    def typed(cls, model: type[T]) -> type['Response[T]']:
        return type(cls.__name__, (cls,), {'model': model})

    @functools.singledispatchmethod
    def render(self, content: pydantic.BaseModel | dict[str, Any] | str) -> bytes:
        if isinstance(content, pydantic.BaseModel):
            content = content.model_dump(mode='json', by_alias=True)
        media_type = self.media_type or self.default_media_type
        match media_type:
            case 'text/html':
                self.media_type = 'text/plain'
                return yaml.safe_dump( # type: ignore
                    data=content,
                    indent=2,
                    default_flow_style=False
                ).encode('utf-8') # type: ignore
            case 'application/yaml':
                return yaml.safe_dump( # type: ignore
                    data=content,
                    default_flow_style=False
                ).encode('utf-8') # type: ignore
            case _:
                return json.dumps(
                    content,
                    ensure_ascii=False,
                    allow_nan=False,
                    separators=(",", ":"),
                ).encode("utf-8")


# Copyright (C) 2023-2024 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Annotated
from typing_extensions import Doc

from canonical.ext.api import ObjectReference
from canonical.ext.api import LocalObjectReference



ResourceUntypedKey = Annotated[
    LocalObjectReference,
    Doc(
        "An identifier for a resource from which the type can not be inferred, "
        "such as an int or a string."
    )
]

ResourceKey = Annotated[
    ObjectReference | ResourceUntypedKey,
    Doc(
        "An identifier for a resource."
    )
]





# Copyright (C) 2023-2024 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .authorizationrequest import AuthorizationRequest
from .authorizationresponse import AuthorizationResponse
from .authorizationstatebuilder import AuthorizationStateBuilder
from .redirectedstate import RedirectedState
from .tokenendpointclient import TokenEndpointClient
from .tokenendpointclient import TokenEndpointClientDepends
from .tokenrequest import TokenRequest
from .tokenrequest import TokenRequestDepends


__all__: list[str] = [
    'AuthorizationRequest',
    'AuthorizationResponse',
    'AuthorizationStateBuilder',
    'RedirectedState',
    'TokenEndpointClient',
    'TokenEndpointClientDepends',
    'TokenRequest',
    'TokenRequestDepends',
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import NotRequired
from typing import TypedDict


class JOSEHeaderDict(TypedDict):
    alg: str
    cty: NotRequired[str] | None
    crit: NotRequired[list[str]]
    jku: NotRequired[str | None]
    #jwk: JSONWebKey | None
    kid: NotRequired[str | None]
    typ: NotRequired[str | None]
    x5c: NotRequired[str | None]
    x5t: NotRequired[str | None]
    #x5t_sha256: str | None
    x5u: NotRequired[str | None]
    claims: NotRequired[dict[str, Any]]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import TypeVar

from canonical.ext.api import APIResourceType
from .replace import Replace
from ..params import ResourceRepository
from ..params import RequestResource


T = TypeVar('T', bound=APIResourceType)


class DefaultReplace(Replace[T]):
    creates = False
    detail = True
    exists = True
    method = 'PUT'
    requires_body = True
    status_code = 205
    verb = 'replace'

    async def handle(
        self,
        repo: ResourceRepository,
        old: RequestResource,
        new: APIResourceType
    ) -> APIResourceType:
        return await self.replace(repo, old, new)

    async def replace(
        self,
        repo: ResourceRepository,
        old: RequestResource,
        new: APIResourceType
    ) -> APIResourceType:
        assert old is not None
        old.replace(new)
        await repo.persist(self.model, old)
        return old


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Generic
from typing import TypeVar
from typing import Union
from typing import TYPE_CHECKING

from canonical.exceptions import ProgrammingError
from ..bases import BaseResource
from ..bases import BaseRootResource
from .rootmodelclass import RootModelBuilder
from .rootmodeldescriptor import RootModelDescriptor
if TYPE_CHECKING:
    from ..apirootresource import APIRootResource

T = TypeVar('T', bound='APIRootResource[Any]')


class APIRootResourceClassBuilder(RootModelBuilder[T], Generic[T]):

    def setup(self, model: type[T], **kwargs: Any):
        super().setup(model, **kwargs)
        self.meta = model.__meta__
        model.__namespaced__ = bool(kwargs.get('namespaced'))

        # Check if all models in the root are
        # implementations of BaseResource.
        for child in self.children:
            if not issubclass(child, BaseResource):
                raise ProgrammingError(
                    "Child classes must inherit from BaseResource."
                )

    def build_create_model(self):
        union = Union[*[x.__create_model__ for x in self.children]] # type: ignore
        return type(self.model.__name__, (BaseRootResource[union],), {
            'model_config': {
                **self.model.model_config,
                'ignored_types': (RootModelDescriptor,)
            },
            'api_version': property(lambda self: self.root.api_version),
            'kind': property(lambda self: self.root.kind),
            'metadata': property(lambda self: self.root.metadata),
            #'replacable': property(lambda self: self.root.replacable),
        })


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .jsonwebkey import JSONWebKey
from .jsonwebkeyset import JSONWebKeySet
from .jws import JWS
from .jwt import JWT


__all__: list[str] = [
    'JSONWebKey',
    'JSONWebKeySet',
    'JWT',
    'JWS',
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

import pydantic

from canonical.ext.resource import Resource
from canonical.ext.resource import TransientMeta


class PermissionQuerySpec(pydantic.BaseModel):
    permissions: set[str] = pydantic.Field(
        default=...,
        description=(
            "The set of permissions that are being requested on a "
            "resource."
        )
    )


class PermissionQuery(Resource[TransientMeta], version='permissionqueries.iam.webiam.io/v1'):
    spec: PermissionQuerySpec = pydantic.Field(
        default=...,
        description=(
            "Specifies the properties of the authorization that "
            "the client wants to know."
        )
    )

    @classmethod
    def factory(cls, spec: dict[str, Any]):
        return cls.model_validate({
            'metadata': {},
            'spec': spec
        })


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .basedatastorestorage import BaseDatastoreStorage
from .resourcerepository import ResourceRepository
from .googleresourcerepository import GoogleDatastoreResourceRepository
from .serviceaccounttokensigner import ServiceAccountTokenSigner
from .serviceaccounttokenverifier import ServiceAccountTokenVerifier


__all__: list[str] = [
    'BaseDatastoreStorage',
    'GoogleDatastoreResourceRepository',
    'ResourceRepository',
    'ServiceAccountTokenSigner',
    'ServiceAccountTokenVerifier',
]


# Copyright (C) 2023-2024 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Self

import pydantic

from ..apiversioned import APIModel
from ..bases.reference import BaseReference
from ..builders import APIModelClassBuilder
from ..fields import Kind


__all__: list[str] = [
    'TypedLocalObjectReference'
]


class TypedLocalObjectReference(APIModel, BaseReference):
    """A `TypedLocalObjectReference` instance contains enough
    information to let you locate the typed referenced object
    inside the same namespace.
    """
    api_group: str = pydantic.Field(
        default=...,
        alias='apiGroup',
        description=(
            "The `apiGroup` field is the group for the resource being "
            "referenced. If `apiGroup` is not specified, the specified "
            "`kind` must be in the core API group. For any other "
            "third-party types, `apiGroup` is required."
        )
    )

    kind: Kind = pydantic.Field(
        default=...,
        description="The type of resource being referenced."
    )

    name: str = pydantic.Field(
        default=...,
        description="The name of resource being referenced."
    )

    @classmethod
    def build_class(
        cls,
        builder: APIModelClassBuilder[Self],
        **kwargs: Any
    ):
        super().build_class(builder, **kwargs)
        builder.set_field_literal('api_group', kwargs.get('api_group'))
        builder.set_field_literal('kind', kwargs.get('kind'))


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import get_args
from typing import Any
from typing import Generic
from typing import TypeVar

import pydantic

from .basemodelclass import BaseModelClassBuilder


T = TypeVar('T', bound=pydantic.RootModel[Any])


class RootModelBuilder(BaseModelClassBuilder[T], Generic[T]):
    children: list[type[pydantic.BaseModel]]

    def setup(self, model: type[T], **kwargs: Any) -> None: # type: ignore
        super().setup(model, **kwargs)
        self.field = model.model_fields['root']
        self.children = list(get_args(self.field.annotation))
        


#!/usr/bin/env python3
#
# Copyright (C) 2019-2020 Cochise Ruhulessin
#
# This file is part of canonical.
#
# canonical is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# canonical is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with canonical.  If not, see <https://www.gnu.org/licenses/>.
import json
import os
import sys
from setuptools import find_namespace_packages
from setuptools import setup

curdir = os.path.abspath(os.path.dirname(__file__))
version = str.strip(open('VERSION').read())
opts = json.loads((open('canonical/package.json').read()))
if os.path.exists(os.path.join(curdir, 'README.md')):
    with open(os.path.join(curdir, 'README.md'), encoding='utf-8') as f:
        opts['long_description'] = f.read()
        opts['long_description_content_type'] = "text/markdown"

setup(
    name='canonical',
    version=version,
    packages=find_namespace_packages(),
    include_package_data=True,
    **opts)



# Copyright (C) 2023-2024 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import cast
from typing import Any
from typing import ClassVar
from typing import Self
from typing import TypeVar

import pydantic

from canonical.ext.crypto import IDataEncryption
from .apiresource import APIResource
from .apiversionedmeta import APIVersionedMeta
from .bases import BaseRootResource
from .builders import APIRootResourceClassBuilder
from .objectmeta import ObjectMeta


T = TypeVar('T', bound=APIResource[Any])


class APIRootResource(BaseRootResource[T]):
    __builder_class__ = APIRootResourceClassBuilder
    __descriptor_class__: ClassVar[type[APIVersionedMeta[Any]]] = APIVersionedMeta
    __meta__: ClassVar[Any]

    @classmethod
    def model_input(cls, data: dict[str, Any]):
        return cls.__create_model__.model_validate(data)

    def __init_subclass__(cls, **kwargs: Any):
        if kwargs:
            kwargs.setdefault('kind', cls.__name__)
            kwargs.setdefault('namespaced', False)
            meta = cls.__descriptor_class__.model_validate({
                **kwargs,
                'root': True
            })
            meta.contribute_to_class(cls, {})
        return super().__init_subclass__()

    @classmethod
    def __pydantic_init_subclass__(cls, **kwargs: Any) -> None:
        if kwargs and not kwargs.get('abstract'):
            cls.__builder__ = cls.__builder_class__(cls, **kwargs)
            cls.build_class(cast(Any, cls.__builder__))
        return super().__pydantic_init_subclass__(**kwargs)

    @classmethod
    def build_class(cls, builder: APIRootResourceClassBuilder[Self]) -> None:
        cls.__create_model__ = builder.build_create_model()

    @classmethod
    def is_namespaced(cls):
        return cls.__namespaced__

    @property
    def api_version(self):
        return self.root.api_version

    @property
    def cache_key(self):
        return self.root.cache_key

    @property
    def key(self):
        return self.root.get_object_ref(self.__meta__).with_model(type(self))

    @property
    def kind(self) -> str:
        return self.root.kind

    @property
    def metadata(self) -> ObjectMeta[Any]:
        return self.root.metadata

    @property
    def model(self):
        return type(self)

    def get_namespace(self):
        return self.root.get_namespace()

    def replace(self, instance: pydantic.BaseModel):
        return self.root.replace(instance)

    def update(self, old: Self | None) -> tuple[Self, bool]:
        if old is None:
            return self.model_validate(self.model_dump()), True
        root, changed = self.root.update(old.root)
        return self.model_validate(root.model_dump()), changed

    def decrypt(self, dek: IDataEncryption[Any]):
        return self.root.decrypt(dek)

    def encrypt(self, dek: IDataEncryption[Any]):
        return self.root.encrypt(dek)


# Copyright (C) 2023-2024 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Literal

import pydantic

from .apimodel import APIModel


__all__: list[str] = [
    'LabelSelectorRequirement'
]


class LabelSelectorRequirement(APIModel):
    key: str = pydantic.Field(
        default=...,
        description="The label that the selector applies to."
    )

    operator: Literal['In', 'NotIn', 'Exists', 'DoesNotExist'] = pydantic.Field(
        default=...,
        description=(
            "The key\'s relationship to a set of values. Valid operators "
            "are `In`, `NotIn`, `Exists` and `DoesNotExist`."
        )
    )

    values: list[str] = pydantic.Field(
        default=...,
        description=(
            "An array of string values. If the operator is `In` or `NotIn` "
            ", the values array must be non-empty. If the operator is "
            "`Exists` or `DoesNotExist`, the values array must be empty."
        )
    )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import asyncio
import logging
import inspect
from typing import cast
from typing import Callable
from typing import TypeVar
from typing import Unpack

import httpx
import pydantic
from httpx._types import URLTypes

from canonical.lib.utils.http import MediaTypeSelector
from .asyncclientparams import AsyncClientParams
from .asyncrequestparams import AsyncRequestParams
from .response import Response


__all__: list[str] = [
    'AsyncClient',
    'AsyncClientParams',
]


T = TypeVar('T')


class AsyncClient(httpx.AsyncClient):
    logger: logging.Logger = logging.getLogger('headless')

    @property
    def domain(self) -> str:
        return self.base_url.netloc.decode()

    def can_retry_timeout(self, request: httpx.Request):
        return request.method in {'GET', 'HEAD', 'OPTIONS'}

    async def request(
        self,
        method: str,
        url: URLTypes,
        *,
        response_model: type[T] | Callable[[bytes], T] = bytes,
        **kwargs: Unpack[AsyncRequestParams]
    ) -> Response[T]:
        abort = False
        attempts: int = 0
        counter = -1
        delay: int = 5
        exception = None
        while True:
            attempts += 1
            if counter > 0:
                counter -= 1
            must_log = (attempts % 3 == 0) or (counter != -1)
            try:
                response = await super().request(method, url, **kwargs)
                response.raise_for_status()
                break
            except httpx.ConnectError as e:
                # Assumption is that there is no internet connection.
                if must_log:
                    self.logger.warning("Network issues prevent request (url: %s)", e.request.url)
                await asyncio.sleep(delay)
                continue
            except httpx.ConnectTimeout as e:
                # A connection timeout may occur due to intermediary issues, or
                # a booting application that has not bound its port yet.
                if must_log:
                    self.logger.warning("Connection timeout (url: %s)", e.request.url)
                await asyncio.sleep(delay)
                continue
            except httpx.ReadTimeout as e:
                # A read timeout means that we've succesfully connected, but there
                # was a timeout reading the response. This should only be retried on
                # safe HTTP methods, because we do not know what actually caused the
                # timeout, thus any destructive operations may actually have been
                # completed succesfully.
                self.logger.warning("Caught timeout (url: %s).", e.request.url)
                if not self.can_retry_timeout(e.request):
                    raise
                await asyncio.sleep(delay)
                continue
            except httpx.HTTPStatusError as e:
                if e.response.status_code != 429:
                    response = e.response
                    break
                if must_log:
                    self.logger.warning("Request was rate-limited (url: %s)", e.response.url)
                await asyncio.sleep(delay)
            except httpx.RemoteProtocolError as e:
                # This is an edge condition where we might try to make a
                # request to a server that closed it ports. There are
                # other reasons, so we limit the amount of retries.
                if counter == -1:
                    counter = 3
                if must_log:
                    self.logger.warning("Remote protocol violation (url: %s)", e.request.url)
                exception = e
                await asyncio.sleep(delay)
            if counter == 0:
                abort = True
            if abort and exception:
                raise exception

        selector = MediaTypeSelector({'application/json'})
        if response.status_code >= 200:
            media_type = selector.select(response.headers.get('Content-Type'))
            if media_type == 'application/json':
                setattr(response, 'result', response.content)
                if inspect.isclass(response_model)\
                and issubclass(response_model, pydantic.BaseModel):
                    setattr(response, 'result', response_model.model_validate(response.json()))

        return cast(Response[T], response)


# Copyright (C) 2023-2024 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Literal

import pydantic

from .apiversioned import APIModel
from .basereference import BaseReference
from .fields import Kind
from .fields import Namespace


__all__: list[str] = [
    'TypedObjectReference'
]


class TypedObjectReference(APIModel, BaseReference):
    """A `TypedObjectReference` instance contains enough information to let
    you inspect or modify the referred object of the given type.
    """
    api_group: str = pydantic.Field(
        default=...,
        alias='apiGroup',
        description=(
            "The `apiGroup` field is the group for the resource being "
            "referenced. If `apiGroup` is not specified, the specified "
            "`kind` must be in the core API group. For any other "
            "third-party types, `apiGroup` is required."
        )
    )

    kind: Kind = pydantic.Field(
        default=...,
        description="The type of resource being referenced."
    )

    name: str = pydantic.Field(
        default=...,
        description="The name of resource being referenced."
    )

    namespace: Namespace = pydantic.Field(
        default='',
        description=(
            "The namespace of resource being referenced, or an "
            "empty string for cluster resources."
        )
    )

    def __init_subclass__(cls, **kwargs: Any):
        return super().__init_subclass__()

    @classmethod
    def __pydantic_init_subclass__(
        cls,
        *,
        namespaced: bool | None = None,
        api_group: str | None = None,
        kind: str | None = None,
        **kwargs: Any,
    ) -> None:
        super().__pydantic_init_subclass__(**kwargs)
        fields = cls.model_fields
        if api_group is not None:
            fields['api_group'].annotation = Literal[f'{api_group}']
        if kind is not None:
            fields['kind'].annotation = Literal[f'{kind}']
        if not namespaced:
            cls.model_fields['namespace'].annotation = Literal['']
            del cls.model_fields['namespace']
        assert cls.model_rebuild(force=True)

    def attach_to_namespace(self, namespace: str) -> None:
        pass


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic

from canonical.lib.types import Base64


class BaseEncryptionResult(pydantic.BaseModel):
    alg: str = pydantic.Field(
        default=...,
        title="Algorithm",
        description="The algorithm used to perform the encryption function."
    )

    kid: str | None = pydantic.Field(
        default=None,
        title="Key Identifier",
        description=(
            "Identifies the key used to perform the encryption function."
        )
    )

    ct: Base64 = pydantic.Field(
        default=...,
        title="Ciphertext",
        description=(
            "The Base64-encoded encryption result."
        )
    )

    def __bytes__(self) -> bytes:
        return self.ct


# Copyright (C) 2023-2024 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import get_args
from typing import Any
from typing import ClassVar
from typing import Generic
from typing import Literal
from typing import Self
from typing import TypeAlias
from typing import TypeVar

import pydantic

from canonical.ext.api import ObjectReference
from .apiresourcelist import APIResourceList
from .listbase import ListBase
from .resource import Resource
from .resourcemeta import ResourceMeta

__all__: list[str] = [
    'RootResource'
]

T = TypeVar('T', bound=Resource[Any])
S = TypeVar('S', bound='RootResource[Any]')


class RootResource(pydantic.RootModel[T], Generic[T]):
    __version__: ClassVar[str]

    _is_namespaced: bool
    group: ClassVar[str]
    base_path: ClassVar[str]
    plural: ClassVar[str]
    List: ClassVar[type[ListBase[Any, Any]]]
    InputModel: ClassVar[type[Self]]
    __meta__: ClassVar[ResourceMeta]

    @property
    def api_version(self):
        return self.root.api_version

    @property
    def key(self):
        ref = ObjectReference(
            api_version=self.api_version,
            kind=self.kind,
            name=self.metadata.name,
            namespace=self.metadata.get_namespace() or '',
            resource_version=self.metadata.resource_version or '',
            uid=self.metadata.uid
        )
        return ref.attach(self.metadata).with_model(type(self))

    @property
    def kind(self):
        return self.root.kind

    @property
    def metadata(self):
        return self.root.metadata

    @property
    def relname(self):
        return self.root.relname

    @classmethod
    def register(cls, resources: APIResourceList):
        return resources.add(cls.__meta__)

    def __init_subclass__(
        cls,
        version: str | None = None,
        namespaced: bool | None = None,
        **kwargs: Any
    ):
        super().__init_subclass__()
        paths: set[str] = set()
        types: tuple[type[Resource[Any]]] = get_args(cls.model_fields['root'].annotation) # type: ignore
        if version:
            assert len(types)
            cls._is_namespaced = types[0].is_namespaced()
            meta = ResourceMeta.fromqualname(cls, version)
            for t in types:
                t.add_to_root(cls, meta)
            meta.contribute_to_class(cls, {}, root=types)
            return
        for model in types:
            paths.add(model.base_path)
        if len(paths) > 1:
            raise ValueError(f"All root models must have the same base path.")
        if paths:
            assert len(types) > 1
            cls.base_path = types[0].base_path
            cls.group = types[0].group
            cls.plural = types[0].plural
            cls._is_namespaced = types[0].is_namespaced()
            cls.KeyType = PrimaryKey.typed(cls) # type: ignore
            cls.__meta__ = ResourceMeta.model_validate({
                **types[0].__meta__.model_dump(),
                'kind': cls.__name__
            })
        cls.List = type(f'{cls.__name__}List', (ListBase[Literal[f'{cls.__name__}List'], cls],), {
            'items': pydantic.Field(
                default_factory=list,
                description=(
                    "The `items` member contains an array "
                    f"of `{cls.__name__}` objects."
                )
            ),
            '__annotations__': {'items': list[cls]}
        })
        cls.List.model_fields['kind'].default = f'{cls.__name__}List'
        cls.List.model_rebuild()

        cls.model_config.update({
            'title': cls.__name__
        })
        cls.model_rebuild()

    @classmethod
    def is_destroyable(cls) -> bool:
        return True

    @classmethod
    def is_namespaced(cls) -> bool:
        return cls._is_namespaced

    @classmethod
    def is_purgable(cls) -> bool:
        return True

    @classmethod
    def model_validate_input(cls, data: dict[str, Any]) -> Self:
        return cls.model_validate(data)

    def can_change(self, old: T) -> bool:
        return self.root.can_change(old)

    def get_comparison_fields(self) -> set[str]:
        return self.root.get_comparison_fields()

    def get_namespace(self) -> str | None:
        return self.root.get_namespace()

    def has_state(self):
        return self.root.has_state()

    def in_namespace(self, namespace: str | None):
        return self.root.in_namespace(namespace)

    def is_consistent(self) -> bool:
        return self.root.is_consistent()

    def model_dump_yaml(self, **kwargs: Any):
        return self.root.model_dump_yaml(**kwargs)

    def replacable(self) -> bool:
        return self.root.replacable()

    async def persist(self, *args: Any, **kwargs: Any):
        obj = await self.root.persist_key(self.key, *args, **kwargs)

        # Ensure that the same type is returned.
        return self.model_validate(obj.model_dump())


ResourceType: TypeAlias = Resource[Any] | RootResource[Any]
ResourceTypeVar = TypeVar('ResourceTypeVar', bound=ResourceType)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
"""Declares :class:`ServerMetadata`."""
import httpx
import pydantic

from canonical.lib.types import HTTPResourceLocator


class ServerMetadata(pydantic.BaseModel):
    model_config = {'populate_by_name': True}

    issuer: HTTPResourceLocator = pydantic.Field(
        default=...,
        title="Issuer",
        description="Authorization server's issuer identifier URL.",
    )

    authorization_endpoint: HTTPResourceLocator | None = pydantic.Field(
        default=None,
        title="Authorization endpoint",
        alias='authorizationEndpoint',
        description="URL of the authorization server's authorization endpoint.",
    )

    token_endpoint: str = pydantic.Field(
        default=...,
        title="Token endpoint",
        alias='tokenEndpoint',
        description="URL of the authorization server's token endpoint.",
    )

    jwks_uri: str | None = pydantic.Field(
        default=None,
        title="JSON Web Key Set (JWKS) URI",
        alias='jwksUri',
        description="URL of the authorization server's JWK Set document.",
    )

    registration_endpoint: str | None = pydantic.Field(
        default=None,
        title="Registration endpoint",
        alias='registrationEndpoint',
        description=(
            "URL of the authorization server's **OAuth 2.0 Dynamic Client "
            "Registration Endpoint**."
        ),
    )

    scopes_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported scopes",
        description=(
            "JSON array containing a list of the OAuth 2.0 `scope` values that "
            "this authorization server supports."
        ),
    )

    response_types_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported response types",
        description=(
            "JSON array containing a list of the OAuth 2.0 `response_type` "
            "values that this authorization server supports."
        ),
    )

    response_modes_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported response modes",
        description=(
            "JSON array containing a list of the OAuth 2.0 `response_mode` "
            "values that this authorization server supports."
        ),
    )

    grant_types_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported grant types",
        description=(
            "JSON array containing a list of the OAuth 2.0 `grant_types` "
            "values that this authorization server supports."
        ),
    )

    token_endpoint_auth_methods_supported: list[str] = pydantic.Field(
        default=[],
        title="Supported client authentication methods",
        description=(
            "JSON array containing a list of client authentication methods "
            "supported by this token endpoint."
        ),
    )

    token_endpoint_auth_signing_alg_values_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported signature algorithms",
        description=(
            "JSON array containing a list of the JWS signing algorithms "
            "supported by the token endpoint for the signature on the JWT "
            "used to authenticate the client at the token endpoint."
        ),
    )

    service_documentation: str | None = pydantic.Field(
        default=None,
        title="Documentation",
        description=(
            "URL of a page containing human-readable information that "
            "developers might want or need to know when using the "
            "authorization server."
        ),
    )

    ui_locales_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported locals for UI",
        description=(
            "Languages and scripts supported for the user interface, "
            "represented as a JSON array of language tag values from BCP 47."
        ),
    )

    op_policy_uri: str | None= pydantic.Field(
        default=None,
        title="Data policy URL",
        description=(
            "URL that the authorization server provides to the person "
            "registering the client to read about the authorization server's "
            "requirements on how the client can use the data provided by the "
            "authorization server."
        ),
    )

    op_tos_uri: str | None= pydantic.Field(
        default=None,
        title="Terms of service URL",
        description=(
            "URL that the authorization server provides to the person "
            "registering the client to read about the authorization server's "
            "terms of service."
        ),
    )

    revocation_endpoint: str | None = pydantic.Field(
        default=None,
        title="Revocation endpoint",
        description="URL of the authorization server's revocation endpoint.",
    )

    revocation_endpoint_auth_methods_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported authentication methods",
        description=(
            "JSON array containing a list of client authentication methods "
            "supported by this revocation endpoint."
        ),
    )

    revocation_endpoint_auth_signing_alg_values_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported signature algorithms",
        description=(
            "JSON array containing a list of the JWS signing algorithms "
            "supported by the revocation endpoint for the signature on the JWT "
            "used to authenticate the client at the revocation endpoint."
        ),
    )

    introspection_endpoint: str | None = pydantic.Field(
        default=None,
        title="Introspection endpoint",
        description="URL of the authorization server's introspection endpoint.",
    )

    introspection_endpoint_auth_methods_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported authentication methods",
        description=(
            "JSON array containing a list of client authentication methods "
            "supported by this introspection endpoint."
        ),
    )

    introspection_endpoint_auth_signing_alg_values_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported signature algorithms",
        description=(
            "JSON array containing a list of the JWS signing algorithms "
            "supported by the introspection endpoint for the signature on the JWT "
            "used to authenticate the client at the introspection endpoint."
        ),
    )

    signed_metadata: str | None = pydantic.Field(
        default=None,
        title="Signed metadata",
        description=(
            "Signed JWT containing metadata values about the authorization "
            "server as claims."
        )
    )

    device_authorization_endpoint: str | None = pydantic.Field(
        default=None,
        title="Device authorization endpoint",
        description=(
            "URL of the authorization server's device authorization endpoint."
        ),
    )

    tls_client_certificate_bound_access_tokens: bool | None = pydantic.Field(
        default=True,
        title="Supports mTLS certificate-bound access tokens",
        description=(
            "Indicates authorization server support for mutual-TLS client "
            "certificate-bound access tokens."
        ),
    )

    mtls_endpoint_aliases: dict[str, str] | None = pydantic.Field(
        default={},
        title="Alternative mTLS endpoints",
        description=(
            "JSON object containing alternative authorization server "
            "endpoints, which a client intending to do mutual TLS will "
            "use in preference to the conventional endpoints."
        ),
    )

    nfv_token_signing_alg_values_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported signature algorithms",
        description=(
            "JSON array containing a list of the JWS signing algorithms "
            "supported by the server for signing the JWT used as NFV Token."
        ),
    )

    nfv_token_encryption_alg_values_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported encryption algorithms",
        description=(
            "JSON array containing a list of the JWE encryption algorithms "
            "(`alg` values) supported by the server to encode the JWT used as "
            "NFV Token."
        ),
    )

    nfv_token_encryption_enc_values_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported content encryption algorithms",
        description=(
            "JSON array containing a list of the JWE encryption algorithms "
            "(`enc` values) supported by the server to encode the JWT used as "
            "NFV Token."
        ),
    )

    userinfo_endpoint: str | None = pydantic.Field(
        default=None,
        title="UserInfo endpoint",
        description="URL of the authorization servers' UserInfo Endpoint.",
    )

    acr_values_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported ACR",
        description=(
            "JSON array containing a list of the Authentication Context Class "
            "References that this authorization server supports."
        ),
    )

    subject_types_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported subject types",
        description=(
            "JSON array containing a list of the Subject Identifier types that "
            "this authorization server supports"
        ),
    )

    id_token_signing_alg_values_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported signature algorithms",
        description=(
            "JSON array containing a list of the JWS signing algorithms "
            "supported by the server for signing the JWT used as ID Token."
        ),
    )

    id_token_encryption_alg_values_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported encryption algorithms",
        description=(
            "JSON array containing a list of the JWE encryption algorithms "
            "(`alg` values) supported by the server to encode the JWT used as "
            "ID Token."
        ),
    )

    id_token_encryption_enc_values_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported content encryption algorithms",
        description=(
            "JSON array containing a list of the JWE encryption algorithms "
            "(`enc` values) supported by the server to encode the JWT used as "
            "ID Token."
        ),
    )

    userinfo_signing_alg_values_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported signature algorithms",
        description=(
            "JSON array containing a list of the JWS signing algorithms "
            "supported by the server for signing the JWT used as UserInfo Endpoint."
        ),
    )

    userinfo_encryption_alg_values_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported encryption algorithms",
        description=(
            "JSON array containing a list of the JWE encryption algorithms "
            "(`alg` values) supported by the server to encode the JWT used as "
            "UserInfo Endpoint."
        ),
    )

    userinfo_encryption_enc_values_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported content encryption algorithms",
        description=(
            "JSON array containing a list of the JWE encryption algorithms "
            "(`enc` values) supported by the server to encode the JWT used as "
            "UserInfo Endpoint."
        ),
    )

    request_object_signing_alg_values_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported signature algorithms",
        description=(
            "JSON array containing a list of the JWS signing algorithms "
            "supported by the server for signing the JWT used as Request Object."
        ),
    )

    request_object_encryption_alg_values_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported encryption algorithms",
        description=(
            "JSON array containing a list of the JWE encryption algorithms "
            "(`alg` values) supported by the server to encode the JWT used as "
            "Request Object."
        ),
    )

    request_object_encryption_enc_values_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported content encryption algorithms",
        description=(
            "JSON array containing a list of the JWE encryption algorithms "
            "(`enc` values) supported by the server to encode the JWT used as "
            "Request Object."
        ),
    )

    display_values_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported display modes",
        description=(
            "JSON array containing a list of the `display` parameter values "
            "that the OpenID Provider supports."
        ),
    )

    claim_types_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported Claims Types",
        description=(
            "JSON array containing a list of the Claims Types "
            "that the OpenID Provider supports."
        )
    )

    claims_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported Claims Types",
        description=(
            "JSON array containing a list of the Claim Names of the Claims "
            "that the OpenID Provider MAY be able to supply values for."
        )
    )

    claims_locales_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported claim locales",
        description=(
            "Languages and scripts supported for values in Claims being "
            "returned, represented as a JSON array of BCP 47."
        ),
    )

    claims_parameter_supported: bool | None = pydantic.Field(
        default=False,
        title="Supports `claims` parameter?",
        description=(
            "Boolean value specifying whether the OP supports use of the "
            "`claims` parameter."
        ),
    )

    request_parameter_supported: bool | None = pydantic.Field(
        default=False,
        title="Supports `request` parameter?",
        description=(
            "Boolean value specifying whether the OP supports use of the "
            "`request` parameter."
        ),
    )

    request_uri_parameter_supported: bool | None = pydantic.Field(
        default=False,
        title="Supports `request_uri` parameter?",
        description=(
            "Boolean value specifying whether the OP supports use of the "
            "`request_uri` parameter."
        ),
    )

    require_request_uri_registration: bool | None = pydantic.Field(
        default=True,
        title="Requires pre-regiration?",
        description=(
            "Boolean value specifying whether the OP requires any `request_uri` "
            "values used to be pre-registered."
        ),
    )

    require_signed_request_object: bool | None = pydantic.Field(
        default=True,
        title="Requires pre-regiration?",
        description=(
            "Indicates where authorization request needs to be protected as "
            "**Request Object** and provided through either `request` or "
            "`request_uri` parameter."
        ),
    )

    pushed_authorization_request_endpoint: str | None = pydantic.Field(
        default=None,
        title="Pushed Authorization Request (PAR) endpoint",
        description=(
            "URL of the authorization server's pushed authorization request "
            "endpoint."
        ),
    )

    require_pushed_authorization_requests: bool | None = pydantic.Field(
        default=False,
        title="Requires PAR?",
        description=(
            "Indicates whether the authorization server accepts authorization "
            "requests only via PAR."
        ),
    )

    introspection_signing_alg_values_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported signature algorithms",
        description=(
            "JSON array containing a list of algorithms supported by the "
            "authorization server for introspection response signing."
        ),
    )

    introspection_encryption_alg_values_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported encryption algorithms",
        description=(
            "JSON array containing a list of algorithms supported by the "
            "authorization server for introspection response content key "
            "encryption (`alg` value)."
        ),
    )

    introspection_encryption_enc_values_supported: list[str] | None = pydantic.Field(
        default=[],
        title="Supported content encryption algorithms",
        description=(
            "JSON array containing a list of algorithms supported by the "
            "authorization server for introspection response content "
            "encryption (`enc` value)."
        ),
    )

    authorization_response_iss_parameter_supported: bool | None = pydantic.Field(
        default=False,
        title="Supports `iss` parameter in authorization response?",
        description=(
            "Boolean value indicating whether the authorization server "
            "provides the `iss` parameter in the authorization response."
        ),
    )

    authorization_signing_alg_values_supported: list[str] = pydantic.Field(
        default=[],
        title="JARM signature algorithms",
        description=(
            "A JSON array containing a list of the JWS signing algorithms "
            "(`alg` values) supported by the authorization endpoint to "
            "sign the response."
        )
    )

    authorization_encryption_alg_values_supported: list[str] = pydantic.Field(
        default=[],
        title="JARM key wrapping or key agreement algorithms",
        description=(
            "A JSON array containing a list of the JWS signing algorithms "
            "(`alg` values) supported by the authorization endpoint to "
            "encrypt the response."
        )
    )

    authorization_encryption_enc_values_supported: list[str] = pydantic.Field(
        default=[],
        title="JARM encryption algorithms",
        description=(
            "A JSON array containing a list of the JWE encryption algorithms "
            "(`enc` values) supported by the authorization endpoint to "
            "encrypt the response."
        )
    )

    _discovered: bool = pydantic.PrivateAttr(default=False)

    async def discover(self, http: httpx.AsyncClient):
        await self.on_discovered(http)
        return self

    async def on_discovered(
        self,
        http: httpx.AsyncClient
    ) -> None:
        pass


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Literal

import pydantic

from canonical.ext.api.objectmeta import ObjectMeta
from canonical.ext.oauth.models.responses import TokenResponse
from canonical.ext.oauth.types import RedirectURI
from canonical.lib.types import HTTPResourceLocator
from .authorizationstatespec import AuthorizationStateSpec


class ForwardingStateSpec(AuthorizationStateSpec[Literal['forwarding']]):
    forward_url: HTTPResourceLocator = pydantic.Field(
        default=...,
        alias='forwardUrl',
        title="Forward URL",
        description=(
            "The URL to which the user agent must be redirected after "
            "obtaining the grant."
        )
    )

    def get_redirect_uri(self) -> RedirectURI | None:
        raise NotImplementedError

    def get_return_url(self, code: str, iss: str | None = None) -> str:
        return self.forward_url.with_query(state=self.params.state)

    def get_user_agent_redirect_uri(
        self,
        metadata: ObjectMeta[str]
    ) -> str:
        return self.params.authorize(self.authorization_endpoint)


    def on_obtained(self, response: TokenResponse) -> dict[str, Any]:
        raise NotImplementedError


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Literal

import pydantic

from canonical.ext.api import ObjectMeta
from canonical.ext.oauth.models.responses import TokenResponse
from canonical.ext.oauth.types import RedirectURI
from .authorizationstatespec import AuthorizationStateSpec


class ProxyingStateSpec(AuthorizationStateSpec[Literal['proxy']]):
    proxy_redirection_endpoint: RedirectURI = pydantic.Field(
        default=...,
        alias='proxyRedirectionEndpoint',
        title="Proxy Redirection Endpoint",
        description=(
            "The redirection endpoint of the intermediate OAuth 2.x/"
            "OpenID Connect proxy."
        )
    )

    def get_redirect_uri(self) -> RedirectURI | None:
        # For proxied requests, the redirect URI that the
        # authorization server has seen is the redirection
        # endpoint of the proxy, and we always include it.
        return self.proxy_redirection_endpoint

    def get_return_url(
        self,
        code: str,
        iss: str | None = None
    ) -> str:
        uri = self.params.redirect_uri or self.redirect_uri
        assert uri
        return uri.redirect(
            code=code,
            iss=iss,
            state=self.params.state
        )

    def get_user_agent_redirect_uri(
        self,
        metadata: ObjectMeta[str]
    ) -> str:
        return self.params.proxy(
            endpoint=self.authorization_endpoint,
            client_id=self.client_id,
            redirect_uri=str(self.get_redirect_uri()),
            state=metadata.name
        )

    def on_obtained(self, response: TokenResponse) -> dict[str, Any]:
        return {
            'status': 'Issued',
            'access_token': response.access_token,
            'refresh_token': response.refresh_token
        }


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import fastapi
import pydantic
from fastapi.security.base import SecurityBase
from fastapi.openapi.models import SecurityBase as SecurityBaseModel
from fastapi.openapi.models import SecuritySchemeType

from canonical.ext.oauth.models import ClientAssertion
from canonical.ext.oauth.utils import DefaultMediaTypeSelector


class RFC7521(SecurityBase):
    selector = DefaultMediaTypeSelector

    def __init__(
        self,
        scheme_name: str,
        description: str | None = None,
    ):
        self.model = SecurityBaseModel(
            type=SecuritySchemeType.oauth2,
            description=description,
        )
        self.scheme_name = scheme_name or self.__class__.__name__

    async def __call__(self, request: fastapi.Request) -> ClientAssertion | None:
        try:
            assertion = ClientAssertion.model_validate_json(await request.json())
        except pydantic.ValidationError:
            assertion = None
        return assertion


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import datetime
from typing import Generic
from typing import TypeVar

import pydantic


R = TypeVar('R')
S = TypeVar('S')


class Condition(pydantic.BaseModel, Generic[S]):
    """Describes the condition of a :class:`VersionedResource`."""

    timestamp: datetime.datetime = pydantic.Field(
        default=...,
        description=(
            "Specifies the date and time at which the condition "
            "emerged."
        )
    )

    message: str = pydantic.Field(
        default='',
        description=(
            "A human readable message indicating details about "
            "the transition. This may be an empty string."
        )
    )

    observed_generation: int = pydantic.Field(
        default=...,
        alias='observedGeneration',
        description=(
            "Represents the `.metadata.generation` that the "
            "condition was set based upon. For instance, if "
            "`.metadata.generation` is currently 12, but "
            "the `.status.conditions[x].observedGeneration` "
            "is 9, the condition is out of date with respect "
            "to the current state of the instance."
        )
    )

    status: S = pydantic.Field(
        default=...,
        description=(
            "Contains a programmatic identifier indicating the "
            "status of the condition's last transition. "
            "Producers of specific condition types may define "
            "expected values and meanings for this field, "
            "and whether the values are considered a guaranteed "
            "API. This field may not be empty."
        )
    )

    def is_final(self) -> bool:
        return False


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import datetime
import decimal
import functools
import hashlib
import struct
from typing import overload
from typing import Any
from typing import Iterable
from typing import Literal
from typing import Protocol



class IHasher(Protocol):
    def update(self, value: bytes) -> None: ...



@overload
def deephash(obj: Any, *, encode: Literal['hex']) -> str: ...


@overload
def deephash(obj: Any, *, encode: None = None) -> bytes: ...


def deephash(obj: Any, *, encode: Literal['hex'] | None = None) -> bytes | str:
    h = hashlib.sha3_256()
    hasher(obj, h)
    return h.hexdigest() if encode == 'hex' else h.digest()


@functools.singledispatch
def hasher(o: Any, h: Any) -> None:
    raise NotImplementedError(type(o).__name__)


@hasher.register
def deephash_bytes(obj: bytes, h: IHasher):
    h.update(obj)


@hasher.register
def deephash_date(obj: datetime.date, h: IHasher):
    hasher((obj.year, obj.month, obj.day), h)


@hasher.register
def deephash_datetime(obj: datetime.datetime, h: IHasher):
    hasher(obj.timestamp(), h)
    hasher(obj.tzinfo or '', h)


@hasher.register
def deephash_decimal(obj: decimal.Decimal, h: IHasher):
    h.update(str.encode(str(obj)))


@hasher.register
def deephash_dict(obj: dict, h: IHasher): # type: ignore
    key_hashes: Iterable[tuple[Any, str]] = sorted([
        (deephash(k, encode='hex'), k) for k in obj.keys() # type: ignore
    ])
    for hashed_key, key in sorted(key_hashes, key=lambda x: x[0]):
        hasher(hashed_key, h)
        hasher(obj[key], h) # type: ignore


@hasher.register
def deephash_float(obj: float, h: IHasher): # type: ignore
    hasher(struct.pack('<d', obj), h)


@hasher.register
def deephash_int(obj: int, h: IHasher): # type: ignore
    hasher(obj.to_bytes(16, 'big'), h)


def deephash_iterable(obj: Iterable[Any], h: IHasher):
    for item in obj:
        hasher(item, h)


@hasher.register
def deephash_int(obj: int, h: IHasher): # type: ignore
    hasher(obj.to_bytes(16, 'big'), h)


@hasher.register
def deephash_list(obj: list, h: IHasher): # type: ignore
    return deephash_iterable(obj, h) # type: ignore


@hasher.register
def deephash_none(obj: None, h: IHasher):
    hasher(b'', h)


@hasher.register
def deephash_set(obj: set, h: IHasher): # type: ignore
    return deephash_iterable(obj, h) # type: ignore


@hasher.register
def deephash_set(obj: tuple, h: IHasher): # type: ignore
    return deephash_iterable(obj, h) # type: ignore


@hasher.register
def deephash_string(obj: str, h: IHasher):
    hasher(str.encode(obj, 'utf-8'), h)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Literal


AuthenticationMethod = Literal[
    'none',
    'client_secret_post',
    'client_secret_basic',
    'client_secret_jwt',
    'private_key_jwt',
    'tls_client_auth',
    'self_signed_tls_client_auth'
]


# Copyright (C) 2023-2024 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Generic
from typing import TypeVar

import pydantic

from .objectreferencemodel import ObjectReferenceModel


__all__: list[str] = [
    'LocalReference'
]

T = TypeVar('T')


class LocalReference(ObjectReferenceModel[T], Generic[T]):
    """A reference to a local object (in the same namespace)."""
    _namespace: str | None = pydantic.PrivateAttr(default=None)

    def is_local(self) -> bool:
        return True

    def get_namespace(self) -> str | None:
        assert self._namespace is not None
        return self._namespace

    def is_namespaced(self) -> bool:
        return True

    def with_namespace(self, namespace: str):
        self._namespace = namespace
        return self


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import decimal
from typing import Generic
from typing import TypeVar

import pydantic

from canonical import Text
from canonical import ResourceName

from .uom import UOM


K = TypeVar('K')
T = TypeVar('T')


class DocumentProduct(pydantic.BaseModel, Generic[K]):
    """Represents a product that is involved in a :term:`Document`,
    such as a request, quote, order or invoice.
    """

    description: Text = pydantic.Field(
        default=...,
        description=(
            "A description of the item, such as the product name."
        )
    )

    labels: dict[str, str] = pydantic.Field(
        default_factory=dict,
        description=(
            "Map of string keys and values that can be used to "
            "organize and categorize (scope and select) objects."
        )
    )

    key: K = pydantic.Field(
        default=...,
        description="Internally identifies the document item."
    )

    origin: ResourceName = pydantic.Field(
        default=...,
        description=(
            "A resource name specifying the origin of the item, identifying "
            "the system that produced it."
        )
    )

    quantity: decimal.Decimal = pydantic.Field(
        default=decimal.Decimal(0),
        description=(
            "The quantity of the product involved for this item."
        )
    )
    
    sku: str = pydantic.Field(
        default=...,
        description=(
            "The Stock Keeping Unit (SKU) that globally identifies "
            "the product."
        )
    )

    uom: UOM = pydantic.Field(
        default_factory=UOM.each,
        description=(
            "The unit of measurement that measures the specified "
            "`quantity`."
        )
    )

    def label(self, type: type[T], name: str) -> T:
        adapter = pydantic.TypeAdapter(type)
        return adapter.validate_python(self.labels.get(name))


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any, Literal

import pydantic

from .resourcemeta import ResourceMeta


class APIResource(pydantic.BaseModel):
    name: str
    namespaced: bool
    kind: str
    verbs: list[str] = []

    def add(self, verb: str) -> None:
        self.verbs = list(sorted({verb, *self.verbs}))

    def path(
        self,
        api_version: str,
        name: str | int | None = None,
        namespace: str | None = None
    ) -> str:
        url = api_version
        if self.namespaced:
            if namespace is None:
                raise TypeError(
                    "The `namespace` parameter can not be None."
                )
            url = f'{url}/namespaces/{namespace}'
        url = f'{url}/{self.name}'
        if name is not None:
            url = f'{url}/{name}'
        return url


class APIResourceList(pydantic.BaseModel):
    kind: Literal['APIResourceList'] = 'APIResourceList'
    groupVersion: str
    resources: list[APIResource] = []
    _index: dict[str, APIResource] = pydantic.PrivateAttr(default_factory=dict)

    def model_post_init(self, _: Any) -> None:
        self._index = {
            resource.kind: resource
            for resource in self.resources
        }

    def add(self, meta: ResourceMeta, plural: str | None = None) -> APIResource:
        plural = plural or meta.plural
        resource = self._index.get(meta.kind)
        if resource is None:
            resource = APIResource(
                name=plural,
                namespaced=meta.namespaced,
                kind=meta.kind,
            )
            self._index[meta.kind] = resource
            self.resources = list(sorted([resource, *self.resources], key=lambda x: x.kind))
        return resource

    def get(self, kind: str) -> APIResource:
        if kind not in self._index:
            raise LookupError(
                f"Unknown resource: {kind}. Supported resources for this group "
                f"are: {str.join(', ', sorted(self._index.keys()))}"
            )
        return self._index[kind]

    def path(self, group: str) -> str:
        return f'{group}/{self.groupVersion}' if group else self.groupVersion